# I2C_DMAC
Arduino Zero (SAMD21/SAMD51) based, non-blocking, I2C library using the Direct Memory Access Controller (DMAC). 

This library employs the SAMD21/SAMD51's DMAC module to transfer data to and from I2C sensors or EEPROM, while simultaneously freeing up the CPU to perform other calculations concurrently. In time critical applications this can lead to a significant boost in processor performance. 

This library was originally designed to significantly increase the performance of the Arduino compatible, I2C sensor based Falcon 1 and 2 multi-rotor (drone) flight controllers:

<img src="https://static.rcgroups.net/forums/attachments/6/0/1/5/8/7/a9606835-172-FalconDiagSmall.png" alt="Falcon Flight Controllers" width="320" height="240"><img src="https://static.rcgroups.net/forums/attachments/6/0/1/5/8/7/a9606837-83-Falcon2Setup.png" alt="Falcon 2 Flight Controller" width="320" height="240">

The current I2C_DMAC Version 1.2.0 is designed to interoperate with other SAMD21/SAMD51 DMAC libraries that have emerged in recent years.

---

## __Contents__

1. [Version](#version)
2. [Arduino Compatibility](#arduino_compatibility)
3. [Installation](#installation)
4. [Usage](#usage)
	1. [I2C_DMAC Library](#i2c_dmac_library)
	2. [Initialisation](#initialisation)
	3. [Reading And Writing](#reading_and_writing)
		1. [Introduction](#introduction)
		2. [1st Layer Functions](#1st_layer_functions)
		3. [2nd Layer Functions](#2nd_layer_functions)
		4. [3rd Layer Functions](#3rd_layer_functions)
		5. [Callback Functions](#callback_functions)
	4. [Code Implementation](#code_implementation)
	5. [DMAC Library Interoperability](#dmac_library_interoperability)
5. [Example Code](#example_code)

<a name="version"></a>
## __Version__

- Version V1.2.0 -- Allow interoperability with other DMAC libraries, add SAMD21E compatibility and update documentation
- Verison V1.1.11 -- Arduino IDE library manager release
- Version V1.1.10 -- Remove I2C instance on writeBusy flag in ReadByte() and ReadBytes() functions
- Version V1.1.9 -- Use default arguments for begin() member function
- Verison V1.1.8 -- Fix bug in Sercom detach callback function and other code optimisations
- Verison V1.1.7 -- Arduino IDE library manager release
- Version V1.1.6 -- Add SERCOM ALT (alternative) peripheral switch for the Metro M4
- Version V1.1.5 -- Activate internal pull-up resistors and increase driver strength
- Version V1.1.4 -- Allow the DMAC to resume normal operation after an early NACK is received
- Version V1.1.3 -- Fixed issue with consecutive calls to writeByte() overwriting data
- Version V1.1.2 -- Allow other classes to simultaneously use remaining DMAC channels
- Version V1.1.1 -- Replaced pinPeripheral() function with port register manipulation
- Version V1.1.0 -- Add Arduino MKR and SAMD51 support, plus multiple I2C instances 
- Version V1.0.0 -- Initial release

<a name="arduino_compatibility"></a>
## __Arduino Compatibility__

- Arduino/Genuino Zero
- Arduino Zero Pro
- Arduino M0 Pro
- Arduino M0
- Arduino MKR Series
- Adafruit M0 and M4 Boards (Metro/Feather/ItsyBitsy)
- Support for smaller SAMD21E variants (Adafruit Qt Py M0)

<a name="installation"></a>
## __Installation__

The I2C_DMAC library can be installed using the Arduino IDE's Library Manager. To access the Library Manager, in the Arduino IDE's menu select _Sketch->Include Library->Manage Libraries...._ In the Library Manager's search bar type I2C_DMAC then select the _"Install"_ button in the I2C_DMAC entry.

Alternatively simply download I2C_DMAC from this Github repository, un-zip or extract the files and place the I2C_DAMC directory in your .../Arduino/libraries/... folder. The .../Arduino/... folder is the one where your Arduino IDE sketches are usually located.

<a name="installation"></a>
## __Usage__

<a name="i2c_dmac_library"></a>
### __I2C_DMAC Library__

Simply include the I2C_DMAC.h file at the beginning of your sketch:

```
#include <I2C_DMAC.h>
```

---

<a name="initialisation"></a>
### __Initialisation__

The I2C_DMAC object is created (instantiated) automatically and by default uses the board's SDA and SCL pins. The I2C_DMAC object can be called using the predefined I2C prefix. The _begin()_ function initialises the I2C_DMAC library with a default, 100kHz SCL clock rate:

```
I2C.begin();	// Initialise the I2C_DMAC library with a 100kHz clock rate
```

To specify a 400kHz clock rate instead, just specify the number of Hz as the first argument:

```
I2C.begin(400000);	// Initialise the I2C_DMAC library with a 400kHz clock rate
```

Thereafter, the I2C clock rate can be set at any time using the _setClock()_ function:

```
I2C.setClock(100000);	// Set the clock rate back to 100kHz
```

Note that for Adafruit's Metro M4 board it is necessary to specify that the I2C port is using the alternative SERCOM (SERCOM_ALT), as well as declaring either the default 8-bit or 16-bit register address, (although the register address size (8-bit or 16-bit) can be changed at any time, depending on the sensor the I2C_DMAC library is communicating with):

```
I2C.begin(400000, REG_ADDR_8BIT, PIO_SERCOM_ALT);	// Initialise the I2C_DMAC with an alternative SERCOM_ALT setting
```

By default the I2C_DMAC uses channel 0 to write and 1 to read, but it is possible to select your DMAC channels of choice: (0-11 on SAMD21 based boards and 0-31 on the SAMD51). These functions are useful if either multiple instances of the I2C_DMAC library are used (see below), or if the I2C_DMAC library is interoperating with other DMAC libraries and channels 0 and 1 are already being used:

```
I2C.setWriteChannel(10);	// Set the I2C_DMAC's write channel
I2C.setReadChannel(11);		// Set the I2C_DMAC's read channel
```

Another feature allows to DMAC read and write channel priority levels to be set (0 lowest - 3 highest). The default is level 0 (lowest):

```
I2C.setPriority(3); 	// Set the read and write channels priority to 3 (highest)
```

Furthermore, multiple instances of the I2C_DMAC library can be called by using the I2C_DMAC's constructor, specifying the base address of SERCOM (Serial Communication) module to be used, together with the additional I2C port's SDA and SCL Arduino pin numbers: 

```
I2C_DMAC I2C1(&sercom2, 4, 3);   // Create (instantiate) a second I2C1 object with sercom2, on pins D4 (SDA) and D3 (SCL)
```

Note that multiple I2C_DMAC instances must also select unused DMAC channels, not used by the first, default instance.

A demonstration sketch: "MPU6050_Gyroscope_V2.ino", using two MPU6050 gyroscope/accelerometer devices is included in the example code.

I2C_DMAC Version 1.2.0 is capable of operating with other DMA libraries, for more information see [DMAC Library Interoperability](#dmac_library_interoperability).

---

<a name="reading_and_writing"></a>
### __Reading and Writing__

<a name="introduction"></a>
#### __Introduction__

The purpose of the I2C_DMAC library is to offload the duties of I2C communication to the DMAC, thereby freeing the CPU sketch to simultaneously continue with other tasks. After initialising and initiating a data transfer, the CPU can at any time check if the DMAC tranfer is complete, by simply polling the library's read and write busy flags. The library also offers the possiblity of excuting user defined callback functions whenever a DMAC read or write operation completes, but also if a DMAC or SERCOM error occurs.   

The I2C_DMAC library has a set of read and write functions, to cover the myriad of I2C sensor types and includes both 8-bit and 16-bit register address implementations. 

The write functions allow for the transmission of the device address, plus the following options:

- Device Address -> Data -> Data Count (bytes)
- Device Address -> 8-bit Register Address
- Device Address -> 16-bit Register Address
- Device Address -> 8-bit Register Address -> 1 Byte Data
- Device Address -> 8-bit Register Addresss -> Data -> Data Count (bytes)
- Device Address -> 16-bit Register Address -> 1 Byte Data
- Device Address -> 16-bit Register Address -> Data -> Data Count (bytes)

The 8-bit register address is used to access most small I2C ICs, such as sensors, while the 16-bit register address can be used to access I2C EEPROM devices.

The read functions allow for the transmission of the device address, plus the reception of data with the following options:

- Device Address -> 1 Data Byte
- Device Address -> Data -> Data Count (bytes)

Multiple byte reads and writes can be used with I2C devices that auto-increment their internal register address.

Note that the I2C_DMAC doesn't use a ring buffer like the standard Wire library, it simply allows you to send and receive data from memory already allocated in your program. This also makes it more efficient as it isn't necessary to pull data off the ring buffer, the data is instead transfer directly to where you specify.

The I2C_DMAC library has 3 layers of functions that progressively combine DMAC initialisation, read, write and synchronisation operations. The 1st layer functions that perform the individual operations can be used where speed is the overriding factor. The 2nd layer functions combine the DMAC initialisation and read/write, while still keeping the read and write operations separate. The 3rd layer functions combine both read and write DMAC operations into single functions. These 3rd layer functions however still require a separate read synchronisation.

<a name="1st_layer_functions"></a>
#### __1st Layer Functions__

As mentioned previously, the 1st layer functions are the most granular, separating the initialisation, read, write and synchronisation operations. When continuously  communicating with a single I2C device, it allows the DMAC to be initialised once and followed by subsequent mulitple calls to the _read()_ and _write()_ functions. This significantly increases code efficiency, since it saves having to intialise the DMAC every transmission.

To initialise the DMAC to write to set the value of the I2C device's register address:

```
I2C.initWriteRegAddr(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS);	// Initialise a write to an I2C device's register
```

To initialise the DMAC to write single byte of data:

```
uint8_t data = 0x00;
I2C.initWriteByte(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS, data);	// Initialise a single byte write of data to a register
```

To initialise the DMAC to write a block of data requires the address of data, together with the byte count:

```
uint8_t data[6] = {};
I2C.initWriteBytes(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS, data, sizeof(data));	// Initialise a multiple byte write of data to a register
```

To initialise the DMAC to write a block of data to an I2C device without a register address:

```
uint8_t data[6] = {};
I2C.initWriteBytes(DEVICE_ADDRESS, data, sizeof(data));	// Initialise a multiple byte write of data
```

To initialise the DMAC to read a single byte of data:

```
I2C.initReadByte(I2C_DEVICE_ADDRESS);	// Initialise a single byte read of data
```

Note that the I2C_DMAC object stores the returned byte of data itself. After transmission and synchronisation (see below), this can be retrieved with the function _getData()_:

```
uint8_t data = I2C.getData();	// Output the received byte of data
```

To initialise the DMAC read multiple bytes of data:

```
uint8_t data[6] = {};
I2C.initReadBytes(I2C_DEVICE_ADDRESS, data, sizeof(data));	// Initialise the read of multiple bytes of data
```

The actual writing and reading of data is initiated with the _write()_ and _read()_ functions:

```
I2C.write();	// Write out the data
```

```
I2C.read();	// Read in the data
```

Once either of these non-blocking _read()_ or _write()_ functions have been called, the DMAC (rather than the CPU) works in conjunction with the SERCOM module to transfer the data, thereby freeing he CPU's sketch to continue with other tasks.

Code subsequently executed by the CPU happens in parallel to, and is concurrent with, the data transfer. To allow the sketch to check if the DMAC read or write operation is complete, it is necessary to poll the respective write and read busy flags. This effectively synchronises the data transfer with the sketch:

```
while(I2C.writeBusy);	// Wait for a DMAC write channel to complete
```

```
while(I2C.readBusy);	// Wait for a DMAC read channel to complete
```

Using a _while()_ statement as shown above causes the sketch to block until the DMAC transfer has completed. 

__Warning: It is important that the transfer synchronization has completed, before initiating another transfer on the I2C bus.__

The I2C_DMAC is also able to execute interrupt driven callback functions when the DMAC completes, for more information see [Callback Functions](#callback_functions).

<a name="2nd_layer_functions"></a>
#### __2nd Layer Functions__

The 2nd layer functions combine the DMAC initialisation with the _write()_ or _read()_ function. These functions still require subsequent synchronisation with sketch using the read and write busy flags (described in the section above).

To initialise the register address and write a single byte of data:

```
uint8_t data = 0x00;
I2C.writeByte(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS, data);	// Initialise and write a single byte of data to a register
```

To intialise the register address and write a multiple bytes of data:

```
uint8_t data[6] = {};
I2C.writeBytes(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS, data, sizeof(data));	// Initialise and write multiple bytes of data to a register
```

To initialise and write to set the I2C device's register address:

```
I2C.writeRegAddr(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS);		// Initialise and write to set the value of the register address
```

To initialise and read a single byte transfer:

```
I2C.readByte(I2C_DEVICE_ADDRESS); // Initialise and read a single byte of data
```

Note that the I2C_DMAC object stores the returned byte of data itself. After synchronisation, this can be retrieved with the function _getData()_:

```
uint8_t data = I2C.getData();	// Output the received byte of data
```

To initialise and read a multiple byte transfer:

```
uint8_t data = {};
I2C.readBytes(I2C_DEVICE_ADDRESS, data, sizeof(data));	// Initialise and read multiple bytes of data
```

<a name="3rd_layer_functions"></a>
#### __3rd Layer Functions__

The 3rd layer functions combine write and read operations, but still require subsequent read synchronisation, (write synchronisation happens within these functions).

To initialise a write to the I2C device's register address, followed by a single byte read of data:

```
I2C.readByte(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS);	// Initialise a write to the register address, followed by a single byte read of data
```

Note that the I2C_DMAC object stores the returned byte of data itself. After synchronisation, this can be retrieved with the function _getData()_:

```
uint8_t data = I2C.getData();		// Output the received byte of data
```

To intialise a write to the I2C device's register address, followed by a multiple byte read of data:

```
uint8_t data[6] = {};
I2C.readBytes(I2C_DEVICE_ADDRESS, REGISTER_ADDRESS, data, sizeof(data));	// Initialise a write to the register address, followed by a multiple byte read of data
```

<a name="callback_functions"></a>
#### __Callback Functions__

The I2C_DMAC library implements both DMAC and SERCOM callback functions. These allow the user to run interrupt driven custom code whenever either the DMAC write or read operation sucessfully completes, or alternatively when a DMAC or SERCOM error occurs. Note that these callback functions occur within the context of the calling interrupt, therefore it is necessary to define any shared varibles used both in the loop() and callback as _volatile_.

To attach a write or read complete callback function, simply use the respective _attachWriteCallback()_ or _attachReadCallback()_, specifying the callback function to be called, for example:

```
I2C.attachWriteCallback(writeCallbackFunction);	// Attach a write callback function
I2C.attachReadCallback(readCallbackFunction);	// Attach a read callback function

void writeCallbackFunction()	// Write callback function
{
  // Add custom code here...
}

void readCallbackFunction()	// Read callback function
{
  // Add custom code here...
}
```

Furthermore, to add a DMAC or SERCOM error callback use the respective _attachDmacErrorCallback()_ or _attachSercomErrorCallback()_, again specifying the callback function:

```
I2C.attachDmacErrorCallback(callbackFunction);	// Attach a DMAC error callback function
```

```
I2C.attachSercomErrorCallback(callbackFunction);	// Attach a SERCOM error callback function
```

Likewise to detach the callback use the detach functions:

```
I2C.detachWriteCallback();
```

```
I2C.detachReadCallback();
```

```
I2C.detachDmacErrorCallback();
```

```
I2C.detachSercomErrorCallback();
```

The _DMAC_Handler()_ and _SERCOMx_Handler()_ interrupt service routines are provided as weak linker symbols allowing them to be overridden, either in a sketch where they can be directly substitued by the inclusion of user defined _DMAC_Handler()_ and _SERCOMx_Handler()_ functions, or alternatively by another DMAC library.

It is also possible to run the I2C_DMAC library without using any DMAC interrupts at all. This is useful for interoperability with other DMAC libraries. See [DMA Library Interoperability](#dma_library_interoperability).

---

<a name="code_implementation"></a>
### __Code Implementation__

An example of accessing the WHO_AM_I register on the MPU6050 gyroscope/accelerometer using the three levels of functions provided by the I2C_DMAC library:

```#include <I2C_DMAC.h>
#define MPU6050_ADDRESS 0x68                         // Device address when ADO = 0
#define WHO_AM_I        0x75                         // Should return 0x68
void setup() 
{
  SerialUSB.begin(115200);                           // Activate the native USB port
  while(!SerialUSB);                                 // Wait for the native USB to be ready
	  
  I2C.begin(400000);                                 // Start I2C bus at 400kHz
	
  // 3rd Layer Functions: Combined DMAC initialisation, write and read
  I2C.readByte(MPU6050_ADDRESS, WHO_AM_I);           // Read the WHO_AM_I register 
  while(I2C.readBusy);                               // Wait for synchronization  
  SerialUSB.println(I2C.getData(), HEX);             // Output the result
	
  // 2nd Layer Functions: Combined DMAC and write initialisation, followed by read
  I2C.writeRegAddr(MPU6050_ADDRESS, WHO_AM_I);       // Write the register address
  while(I2C.writeBusy);                              // Wait for synchronization
  I2C.readByte(MPU6050_ADDRESS);                     // Read the WHO_AM_I register
  while(I2C.readBusy);                               // Wait for synchronization
  SerialUSB.println(I2C.getData(), HEX);             // Output the result
	
  // 1st Layer Functions: Independent DMAC initialisation, write and read
  I2C.initWriteRegAddr(MPU6050_ADDRESS, WHO_AM_I);   // Initialise the DMAC write
  I2C.initReadByte(MPU6050_ADDRESS);                 // Initialise the DMAC read
  I2C.write();                                       // Write the register address
  while(I2C.writeBusy);                              // Wait for synchronization
  I2C.read();                                        // Read the WHO_AM_I register
  while(I2C.readBusy);                               // Wait for synchronization
  SerialUSB.println(I2C.getData(), HEX);             // Output the result 
}
void loop() {}
```

---

<a name="dmac_library_interoperability"></a>
### __DMAC Library Interoperability__

I2C_DMAC Version 1.2.0 is designed to interoperate with other SAMD21/SAMD51 DMAC libraries, such as Adafruit's revised SPI library.

The I2C_DMAC library now auto detects if another library is also using the DMAC peripheral and responds by referencing the other library's DMAC descriptors instead of its own, (descriptors are a set of DMA instructions held in RAM). It also yields control of the DMAC handler/interrupt service routine functions to the other library as well. 

To auto detect the presence of the other library the I2C_DMAC's _begin()_ function must occur after the other library's, for example:

```
// ...
void setup()
{
  // ...
  otherDMA.begin();	// Initialise the other DMA library	
  I2C.begin();	// The I2C_DMAC initialisation must occur after the other DMA library
  // ...
}
// ...
```

Yielding control of the DMAC handler functions means that the I2C_DMAC library can no longer use callback functions, and requires the substitution of the interrupt driven read and write busy flags with new non-interrupt functions called _isWriteBusy()_ and _isReadBusy()_:

```
while(I2C.isWriteBusy());  // Wait for the DMAC write channel to complete
```

```
while(I2C.isReadBusy());  // Wait for the DMAC read channel to complete
```

Moreover, there is also the feature to turn the I2C_DMAC's interrupt functionality on an off manually, using the _dmacInterruptOn()_ and _dmacInterruptOff()_ functions:

```
I2C.dmacInterruptsOn();		// Enable DMAC interrupts
```

```
IC2.dmacInterruptsOff();	// Disable DMAC interrupts
```

This allows for the possibility that I2C_DMAC library can be used in conjuction with the _isBusyWrite()_ and _isBusyRead()_ functions without using DMAC interrupts, even when the library it is operating independently without other DMAC libraries present.

---

<a name="example_code"></a>
### __Example Code__

The examples directory includes I2C_DMAC example code for the MPU6050 gyroscope/accelerometer device:

- __MPU6050_Gyroscope_V1.ino__ : I2C_DMAC with MPU6050 Gyroscope, Basic Example

- __MPU6050_Gyroscope_V2.ino__ : I2C_DMAC with MPU6050 Gyroscope, Multiple I2C Port Instance with Callback Example

- __MPU6050_Gyroscope_V3.ino__ : I2C_DMAC with MPU6050 Gyroscope, Non-Interrupt Example

- __MPU6050_Temperature_V1.ino__ : I2C_DMAC with MPU6050 Temperature, Basic Example

- __MPU6050_Temperature_V2.ino__ : I2C_DMAC with MPU6050 Temperature, Callback Example

- __MPU6050_WhoAmI_V1.ino__ : I2C_DMAC with MPU6050 "Who Am I" register, Function Layers Example

Note that Adafruit boards use _Serial_ rather than _SerialUSB_ for console communications, just add the following definition to the beginning of the code examples:

```
#define SerialUSB Serial		// Define SerialUSB as Serial
```
